/**
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <stdint.h>
#include <errno.h>
#include <mockturtle-framework.h>

uint32_t msg_seq = 0;
struct rt_application *_app;

/**
 * It get the current time from the internal WRNC timer
 * @param[out] seconds
 * @param[out] cycles
 */
void rt_get_time(uint32_t *seconds, uint32_t *cycles)
{
	*seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	*cycles = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
}


/**
 * It initilizes the declared MQs
 * @param[in] app application declaration
 * @return 0 on success. -1 on error
 */
static inline int trtl_rt_init_mq(void)
{
	int i;

	for (i = 0; i < _app->n_mq; ++i) {
		if (_app->mq[i].flags & TRTL_RT_MQ_FLAGS_OUTPUT) {
			_app->mq[i].buf = mq_map_out_buffer(_app->mq[i].type,
							    _app->mq[i].index);
			mq_purge(_app->mq[i].type, _app->mq[i].index);
			if (_app->mq[i].flags & TRTL_RT_MQ_FLAGS_CLAIM) {
				mq_claim(_app->mq[i].type, _app->mq[i].index);
			}
		} else {
			_app->mq[i].buf = mq_map_in_buffer(_app->mq[i].type,
							   _app->mq[i].index);
		}
	}

	return 0;
}


/**
 * The function initializes the library and it does some compatibility check.
 *
 * Initialization
 * - purge all the output host-message-queue
 *
 * compatibility checks:
 * - check bitstream FPGA ID if compatibile with the application. The value in
 *   app->version.fpga_id will be updated with the FPGA ID from the bitstream
 *
 * @param[in] app application declaration
 * @return 0 on success. -1 on error
 */
int rt_init(struct rt_application *app)
{
	int i, err;
	uint32_t *fpga_id = &app->version.fpga_id; /* in order to overwrite the const */

	_app = app;

	/* Validate actions */
	for (i = 0; i < _app->n_actions; ++i) {
		if (i > __TRTL_MSG_ID_MAX_USER) {
			pr_error("Too many actions defined. Maximum '%d'\n\r",
				  __TRTL_MSG_ID_MAX_USER);
			return -1;
		}
		if (_app->actions[i] == NULL) {
			pr_error("Undefined action %d\n\r", i);
			return -1;
		}
	}

	pr_debug("Running application '%s'\n\r", _app->name);

	/* *fpga_id = 0x1234; /\* TODO read it from the FPGA ID register *\/ */
	for (i = 0; i < app->fpga_id_compat_n; i++) {
		if (app->fpga_id_compat[i] == *fpga_id)
			break;
	}
	if (app->fpga_id_compat_n && app->fpga_id_compat_n == i) {
		pr_error("\tFPGA '0x%"PRIx32"' not compatible with RT app: 0x%"PRIx32" 0x%"PRIx32" (git %"PRIx32")\n\r",
			  _app->version.fpga_id,
			  _app->version.rt_id,
			  _app->version.rt_version,
			  _app->version.git_version);
		return -1;
	}
	pr_debug("\tapp id     \t'0x%"PRIx32"'\n\r",
		 _app->version.rt_id);
	pr_debug("\tapp version\t'%"PRId32".%"PRId32"'\n\r",
		  RT_VERSION_MAJ(_app->version.rt_version),
		  RT_VERSION_MIN(_app->version.rt_version));
	pr_debug("\tsource id  \t'0x%"PRIx32"'\n\r",
		 _app->version.git_version);

	err = trtl_rt_init_mq();
	if (err)
		return err;

#ifdef LIBRT_DEBUG
	pr_debug("Exported Variables\n\r");
	for (i = 0; i < _app->n_variables; ++i)
		pr_debug("[%d] addr %p | mask 0x%"PRIx32" | off %d\n\r",
			  i,
			  _app->variables[i].addr,
			  _app->variables[i].mask,
			  _app->variables[i].offset);
	pr_debug("Exported Buffers\n\r");
	for (i = 0; i < _app->n_buffers; ++i)
		pr_debug("[%d] addr %p | len %"PRId32"\n\r", i,
			  _app->buffers[i].buf,
			  _app->buffers[i].len);
#endif

	return 0;
}
